/*****************************
//Justin Thaler
//May 7, 2011
//Implementation of interactive "bounded" protocol for F0, originally due 
//to "Verifying Computations with Streaming Interactive Proofs"
//by Cormode, Thaler, and Yi.
//Code developed for "Practical Verified Computation with Streaming Interactive Proofs"
//by Cormode, Mitzenmacher, and Thaler.
********************************/

#include <cstdlib>
#include <iostream>
#include <time.h>

#define MASK 4294967295 //2^32-1
#define PRIME 2305843009213693951 //2^61-1

typedef unsigned long long uint64;

using namespace std;

uint64 myPow(uint64 x, int b) 
{
  uint64 i = 1;
  for (int j = 0; j < b; j++)  
     i *= x;
  return i;
}

//efficient modular arithmetic function for p=2^61-1. Only works for this value of p.
//This function might
//return a number slightly greater than p (possibly by an additive factor of 8);
//It'd be cleaner to check if the answer is greater than p
//and if so subtract p, but I don't want to pay that
//efficiency hit here, so the user should just be aware of this.
uint64 myMod(uint64 x)
{
  return (x >> 61) + (x & PRIME);
}  
//efficient modular multiplication function mod 2^61-1
uint64 myModMult(uint64 x, uint64 y)
{
	uint64 hi_x = x >> 32; 
	uint64 hi_y = y >> 32;
	uint64 low_x = x & MASK;
	uint64 low_y = y & MASK;

	//since myMod might return something slightly large than 2^61-1,  
	//we need to multiply by 8 in two pieces to avoid overflow.
	uint64 piece1 = myMod((hi_x * hi_y)<< 3);
	uint64 z = (hi_x * low_y + hi_y * low_x);
	uint64 hi_z = z >> 32;
	uint64 low_z = z & MASK;

	//Note 2^64 mod (2^61-1) is 8
	uint64 piece2 = myMod((hi_z<<3) + myMod((low_z << 32)));
	uint64 piece3 = myMod(low_x * low_y);
	uint64 result = myMod(piece1 + piece2 + piece3);

	return result;
}    


uint64* choose_r(uint64 d)
{
  //choose a random vector of dimension d (this is the random point in the LDE)
  uint64* r = (uint64*) malloc(d * sizeof(uint64));
  for(int i=0; i < d; i++)
  {
     r[i] = myMod(rand()); 
  }  
  return r;
}

uint64 chi(uint64 v, uint64* r, uint64 n)
{ 
  uint64 x=v;
  uint64 c = 1;
  for(uint64 i = 0; i <n; i++)
  {
    if( x&1 )
      c=myModMult(c, r[i]);
    else
      c=myModMult(c, 1+PRIME-r[i]);
    x=x>>1;
  }
  return c;
}

//Computes the binary characteristic vector function evaluated over dimensions
//e to d, based on binary index v (view v as a binary vector even though it's an integer), 
//evaluated on vector r. Note this function actually modified v
uint64 chiPiece(uint64 e, uint64 d, uint64 v, uint64* r)
{
  uint64 c = 1;
  for(uint64 i = e; i <d; i++)
  {
    if( v&1 )
      c=myModMult(c, r[i]);
    else
      c=myModMult(c, (1+ myMod(PRIME-r[i])));
    v=v>>1;
  }
  return c;
}


// builds a look up table of products of chi values
// our table will give values for products of blocks of w variables -- 
// this keeps table small (only d/w rows and 2^w cols)
uint64** makechi(uint64 w, uint64 rd, uint64 d, uint64* r)
{
  uint64 rw=myPow(2, w);
  uint64** chilook = (uint64**) malloc(rd * sizeof(uint64*));
  for(int i=0; i < rd; i++)
	chilook[i]=(uint64*) malloc(rw * sizeof(uint64));
  uint64 lo=0;
  uint64 hi=w;
  for(uint64 j = 0; j < rd; j++)
  { //for each block of w coordinates
    for(uint64 i = 0; i < rw; i++)
    { //for each possible set of values of the w coordinates,
      //evaluate chi_1... chi_{2^w} (viewing each subscript as a w-dimensional vector) 
      //over dimensions lo to hi at location vector r
      chilook[j][i]=chiPiece(lo,hi,i,r);
    }
    lo=hi;
    hi=hi+w;
    if (hi>d) //d is the dimensionality of our view of universe [2]^d
      hi=d; //if last block contains fewer than w variables, make sure we only look at ones that exist
  }
  return chilook;
}

//extrapolate the global vector a of length u to location r (a is frequency vector)
//That is, compute \sum_i a_i chi_i(r)
uint64 extrap(uint64 d, uint64 u, uint64* a, uint64* r)
{
  uint64 f=0;
  uint64 w=8; //block length. Before, we had w=d/2 resulting in table of size 2^(d/2)=sqrt(u), which seems like cheating.
  uint64 rw=myPow(2, w);
  uint64 mask=rw-1; //used to pull off last w bits of numbers
  uint64 rd=d/w;//number of blocks of length w we need to look over in the lookup table
  if(d%w>0)
    rd+=1;
  uint64** chilook=makechi(w,rd, d, r);
  
  //temporary variables used below
  uint64 i =0;
  uint64 j=0;
  uint64 delta;
  uint64 k;
  
  while(i<u)
  {
    delta=a[i];
    k=i;
    j=0;
    //evaluate a_i chi_i(r)
    while(j<rd) 
    { 
      delta=myModMult(delta, chilook[j][k&mask]);
      k=k>>w;
      j+=1;
    }
    f = myMod((f+delta));
    i+=1;
  }
  for(int i = 0; i < rd; i++)
    free(chilook[i]);
  free(chilook);
  return f;
}


//Performs Extended Euclidean Algorithm
//Used for computing multiplicative inverses mod p
//Computes a solution  to u*u1 + p*u2 = gcd(u,v)=u3
//only works for p=2^61-1
void extEuclideanAlg(uint64 u, uint64* u1, uint64* u2, uint64* u3)
{
  *u1 = 1;
  *u2 = 0;
  *u3 = u;
  uint64 v1 = 0;
  uint64 v2 = 1;
  uint64 v3 = PRIME;
  uint64 q;
  uint64 t1;
  uint64 t2;
  uint64 t3;
  do
  {
    q = *u3 / v3;
    //t1 = *u1 + p - q * v1;
    //t2 = *u2 + p - q*v2;
    //t3 = *u3 + p - q*v3;
    t1 = myMod((*u1) + PRIME - myModMult(q, v1));
    t2 = myMod((*u2) + PRIME - myModMult(q, v2));
    t3 = myMod((*u3) + PRIME - myModMult(q, v3));
    (*u1) = v1;
    (*u2) = v2;
    (*u3) = v3;
    v1 = t1;
    v2 = t2;
    v3 = t3;
  }while(v3 != 0 && v3!= PRIME);
}

//Computes the modular multiplicative inverse of a modulo m,
//using the extended Euclidean algorithm
//only works for p=2^61-1
uint64 inv(uint64 a)
{
  uint64 u1;
  uint64 u2;
  uint64 u3;
  extEuclideanAlg(a,&u1, &u2, &u3);
  if(u3==1)
      return myMod(u1);
  else
      return 0;
}

//evaluates chi_0(r) where chi_0 takes 0 to 1 and x \in [1, Fmax] to 0
uint64 chi_zero(uint64 r, uint64 Fmax, uint64 inv_prod)
{
  uint64 mult=inv_prod;
  for(uint64 i = 1; i < Fmax; i++)
    mult = myModMult(mult,PRIME+r-i);
  return mult;
}
   
//extrapolate vector vec of length n to location r
//this is slowm but only used once per round, to extrapolate (Fmax-1)-degree polynomial send by H
uint64 extrap1d(uint64* vec, uint64 n, uint64 r)
{
  uint64 result=0;
  uint64 mult=1;
  for(uint64 i = 0; i < n; i++)
  {
    mult=1;
    for(uint64 j=0; j < n; j++)
    {
      if (i>j)
        mult=myModMult(myModMult(mult, myMod(r-j+PRIME)), inv(i-j) );
      if (i<j)
        mult=myModMult(myModMult(mult, myMod(r-j+PRIME)), inv(myMod(i+PRIME-j)) );
    }
    result=myMod(result+myModMult(mult, vec[i]));
  }
  return result;
}


//update a fast
void update_a_fast(uint64 num_effective, uint64 rjm1, uint64* vals)
{
  uint64 index;
  for(uint64 k=0; k < num_effective; k++)
  {
     index = k >> 1;
     if( k & 1)
       vals[index]= myMod(vals[index] + myModMult(vals[k], rjm1));
     else
       vals[index]=myModMult(vals[k], 1+PRIME-rjm1);
  }
}

//cycles through the rounds, computing H and V's messages and timing how uint64 it takes
double rounds(uint64 d, uint64* r, uint64* a, double* ct_secs, uint64 inv_prod, uint64 Fmax, uint64 final_check)
{
  uint64** y = (uint64**) malloc(d * sizeof(uint64*));
  uint64 check=0;
  uint64 ct=0; //check time
  uint64 k = 0;

  clock_t t = clock();

  uint64 num_effective=myPow(2, d);

  //In a real system, P and V's messages need to be interleaved so
  //P doesn't know r_j until after he commits to g_j. But here, we are
  //just interested in determining how long it takes to compute the messages.

  for(uint64 j=0; j < d; j++) // computes all the messages from prover to verifier
  {
      y[j]=(uint64*) calloc(Fmax, sizeof(uint64));
      for(k = 0; k < (num_effective>>1); k++)
      {
        for(uint64 m = 0; m < Fmax; m++)
        { //a[k] gives f(r0...rj-1, k). Want to plug in m for x_j
	    y[j][m]=myMod(y[j][m] + 1+PRIME-chi_zero(myModMult(a[(k<<1)+1], m)+myModMult(a[(k<<1)], PRIME+1-m), Fmax, 
		inv_prod));
	}
      }
      update_a_fast(num_effective, r[j], a);
      num_effective = num_effective >> 1;
  }
  clock_t pt=clock()-t; //prover time

  t=clock();
  for(uint64 j = 0; j < d; j++) //checks all messages from prover to verifier
  {
    if (j>0)
    {
      if( (check != myMod(y[j][0] + y[j][1])) && (check != myMod(y[j][0] + y[j][1]) + PRIME)
          && (check + PRIME != myMod(y[j][0] + y[j][1])) ) // check consistency of messages
      {
        cout << "Check failed: j is " << j << "check is: ";
	cout << check << "!=" <<  myMod(y[j][0] + y[j][1]);
        cout << "y[" << j << "][0]: " << y[j][0] << " y[" << j << "][1]: " << y[j][1] << endl;
	cout << "y[" << j << "][0]: " << (signed long long) (y[j][0]-PRIME);
	cout  << " y[" << j << "][1]: " << (signed long long) (y[j][1]-PRIME) << endl; 
	cout << "check " << (signed long long) (check-PRIME) << endl;
      }
    }
    check=extrap1d(y[j], Fmax, r[j]); //compute next check value at random location r
  }

  if((check != final_check) && (check+PRIME != final_check) && (check != final_check + PRIME))
  {
 	cout << "Final check failed!" << endl;
        cout << "check: " << check << "final_check: " << final_check << endl;
        cout << "dif is: " << myMod(check+PRIME-final_check) << endl;
  }
  else
    cout << "Final check passed." << endl;

  *ct_secs = ((double) ct)/CLOCKS_PER_SEC;
  
  for(int j = 0; j < d; j++)
    free(y[j]);
  free(y); 
  
  return ((double) pt)/CLOCKS_PER_SEC; 
}

int main(int argc, char** argv)
{  
  std::cout.setf(ios::fixed, ios::floatfield);
  std::cout.setf(ios::showpoint);
  std::cout.precision(10);

  uint64 d; 

  if(argc < 3) 
  {
    cout << "not enough command line args\n";
    cout << "Please provide log(universe size) followed by Fmax \n";
    exit(1);
  }

  d=(uint64) atoi(argv[1]); //read dimensionality from command line
  uint64 u=myPow(2, d); //universe size

  //Note Fmax here is really supposed to be Fmax-1. The wrong answer will be given if you do not supply a strict
  //upper bound on Fmax here.
  uint64 Fmax=atoi(argv[2]);
  
  uint64* a = (uint64*) malloc(u*sizeof(uint64));
  if(a == NULL)
  {
     cout << "must not have enough memory; a is NULL" << endl;
     exit(1);
  }
  uint64 i=0;
  srand(time(NULL));
  uint64 F0=0;
  while(i<u) // randomly create a vector of values
  {
    a[i] = rand() % (Fmax-1); 
    if(a[i] != 0)
    	F0++;
    i+=1;
  }

  cout << "Actual F0 is: " << F0 << endl;

  uint64* r=choose_r(d);
  clock_t t=clock();

  //First, prover and verifier are going to compute a product table of inverses from (0-1) to (0-Fmax) because 
  //he will need to reuse them to compute x_0(r)
  uint64 inv_prod = 1;
  for(i=1; i < Fmax; i++)
  {
    inv_prod = myModMult(inv_prod, inv(PRIME-i));
  }

  uint64 fr=extrap(d,u, a, r); //evaluate f extrapolated to random location r

  uint64 final_check = myMod(1+PRIME-chi_zero(fr, Fmax, inv_prod)); 

  clock_t vt=clock()-t;

  double vt_secs =((double) vt)/CLOCKS_PER_SEC;

  double ct_secs;

  double pt_secs=rounds(d, r, a, &ct_secs, inv_prod, Fmax, final_check);
   
  cout << "N\tVerifT\tProveT\tCheckT\tVerifS\tProofS \n";
  cout << u << "\t" << vt_secs << "\t" << pt_secs << "\t" << ct_secs << "\t" << d+1 << "\t" << Fmax*d+1 << endl;
  
  //clean up memory before exiting
  free(r);
  free(a);
}
