/*****************************
//Justin Thaler
//May 7, 2011
//Linearization-based F0 protocol from the paper
//"Practical Verified Computation with Streaming Interactive Proofs"
//by Cormode, Mitzenmacher, and Thaler
********************************/

#include <cstdlib>
#include <iostream>
#include <time.h>
#include <stdlib.h>

#define MASK 4294967295 //2^32-1
#define MASK2 18446744073709551615 //2^64-1
#define MASK3 9223372036854775808 //2^63

#define PRIME 2305843009213693951 //2^61-1

typedef unsigned long long uint64;

using namespace std;

//a full_id stores an item identifier, its location in the stream,
//and a partial contribution mcount of this item to P's current message
typedef struct full_id{
  uint64 id;
  uint64 mcount;
  uint64 inv_lin;
} full_id;

//reverses the bits in a 64-bit data type.
//used because P will need to sort full_id's based on the reversal of their bits
static uint64 reverse(uint64 x) {
 uint64 h = 0;
 int i = 0;

 for(i = 0; i < 64; i++) {
  h = (h << 1) + (x & 1); 
  x >>= 1; 
 }

 return h;
}

//used to sort list of identifiers based on low-order bits (i.e. reverse of id)
int compare (const void * a, const void * b)
{
  uint64 an = ((full_id*) a)->id;
  uint64 bn = ((full_id*) b)->id;
  
  an = reverse(an);
  bn=reverse(bn);
  //cout << "an is: " << an << " bn is : " << bn << " an > bn is: " << (an > bn) << endl; 
  if(an > bn)
    return 1;
  return -1;
}


uint64 myPow(uint64 x, int b) 
{
  uint64 i = 1;
  for (int j = 0; j < b; j++)  
      i *= x;
  return i;
}

//efficient modular arithmetic function for p=2^61-1. Only works for this value of p.
//This function might
//return a number slightly greater than p (possibly by an additive factor of 8);
//It'd be cleaner to check if the answer is greater than p
//and if so subtract p, but I don't want to pay that
//efficiency hit here, so the user should just be aware of this.
uint64 myMod(uint64 x)
{
  return (x >> 61) + (x & PRIME);
}  

//efficient modular multiplication function mod 2^61-1
uint64 myModMult(uint64 x, uint64 y)
{
	uint64 hi_x = x >> 32; 
	uint64 hi_y = y >> 32;
	uint64 low_x = x & MASK;
	uint64 low_y = y & MASK;

	//since myMod might return something slightly large than 2^61-1,  
	//we need to multiply by 8 in two pieces to avoid overflow.
	uint64 piece1 = myMod((hi_x * hi_y)<< 3);
	uint64 z = (hi_x * low_y + hi_y * low_x);
	uint64 hi_z = z >> 32;
	uint64 low_z = z & MASK;

	//Note 2^64 mod (2^61-1) is 8
	uint64 piece2 = myMod((hi_z<<3) + myMod((low_z << 32)));
	uint64 piece3 = myMod(low_x * low_y);
	uint64 result = myMod(piece1 + piece2 + piece3);

	return result;
}    

uint64* choose_r(uint64 num_rounds)
{
  //choose a random vector of dimension num_rounds
  //the last 2*d will represent the random point in the LDE of 
  //the input (viewed as a u \times u matrix) to be evaluated
  uint64* r = (uint64*) calloc(num_rounds, sizeof(uint64));

  for(int i=0; i < num_rounds; i++)
  {
     while(r[i] == 1 || r[i] == 0){
     	r[i] = myMod(rand()); 
     } 
  }  
  return r;
}

//Computes the binary characteristic vector function evaluated over dimensions
//e to d, based on binary index v (view v as a binary vector even though it's an integer), 
//evaluated on vector r. Note this function actually modified v
uint64 chi(uint64 e, uint64 d, uint64 v, uint64* r)
{
  uint64 c = 1;
  for(uint64 i = e; i <d; i++)
  {
    if( v&1 )
      c=myModMult(c, r[i]);
    else
      c=myModMult(c, (1+ myMod(PRIME-r[i])));
    v=v>>1;
  }
  return c;
}

//builds a small lookup table used by V to speed his computation of f(r)
uint64** makechi(uint64 w, uint64 rd, uint64 d, uint64* r)
{
  uint64 rw=myPow(2, w);
  uint64** chilook = (uint64**) malloc(rd * sizeof(uint64*));
  for(int i=0; i < rd; i++)
	chilook[i]=(uint64*) malloc(rw * sizeof(uint64));
  uint64 lo=0;
  uint64 hi=w;
  for(uint64 j = 0; j < rd; j++)
  { //for each block of w coordinates
    for(uint64 i = 0; i < rw; i++)
    { //for each possible set of values of the w coordinates,
      //evaluate chi_1... chi_{2^w} (viewing each subscript as a w-dimensional vector) 
      //over dimensions lo to hi at location vector r
     // cout << "calling chi with i=" << i << "w: " << w << "rd: " << rd << " d: " << d << endl;
      chilook[j][i]=chi(lo,hi,i,r);
    }
    lo=hi;
    hi=hi+w;
    if (hi>d) //d is the dimensionality of our view of universe [2]^d
      hi=d; //if last block contains fewer than w variables, make sure we only look at ones that exist
  }
  return chilook;
}

//extrapolate the global vector a of length u to location r (a is frequency vector)
//That is, compute \sum_i a_i chi_i(r)
uint64** make_lookup(uint64* r, uint64 d)
{
  uint64 w=8; //block length.
  if(w>2*d)
    w=2*d;
  uint64 rw=myPow(2, w);
  uint64 rd=(2*d)/w;//number of blocks of length w we need to look over in the lookup table. 
			//True universe size is 2*d because we are viewing input as matrix here, not as a vector
  if((2*d)%w>0)
    rd+=1;
  return makechi(w,rd, 2*d, r);
}

//update V's hash value. For this protocol, the hash depends on order of updates, which is
//why we don't have V operate on the aggregated frequency vector like in the other protype implementations
//universe size u must be less than 2^32 to avoid overflowing the 64-bit long identifying each item
void update_hash(uint64* hash, uint64** chilook, uint64 rd, uint64 w, uint64 rw, uint64 mask, uint64 identifier)
{
  //assume u and stream length are both equal to 2^d
  //then identifier is 2d bits long (loc, item) where this denotes concatenation
  uint64 j=0;
  uint64 delta=1;
  uint64 k=identifier;

  while(j<rd) 
  { 
      delta=myModMult(delta, chilook[j][k&mask]);
      k=k>>w;
      j+=1;
  }
  *hash = myMod((*hash+delta));
}


//central part of the chi function evaluated at dimension j given binary v
uint64 minchi(uint64 j, uint64 v, uint64 c)
{
  if(((v>>j) & 1)==1) //pull out least significant bit of v and see if it is odd
    return c;
  else
    return myMod((1+ PRIME - c));
}


//Computes a solution  to u*u1 + p*u2 = gcd(u,v)=u3
//only works for p=2^61-1
void extEuclideanAlg(uint64 u, uint64* u1, uint64* u2, uint64* u3)
{
  *u1 = 1;
  *u2 = 0;
  *u3 = u;
  uint64 v1 = 0;
  uint64 v2 = 1;
  uint64 v3 = PRIME;
  uint64 q;
  uint64 t1;
  uint64 t2;
  uint64 t3;
  do
  {
    q = *u3 / v3;
    //t1 = *u1 + PRIME - q * v1;
    //t2 = *u2 + PRIME - q*v2;
    //t3 = *u3 + PRIME - q*v3;
    t1 = myMod((*u1) + PRIME - myModMult(q, v1));
    t2 = myMod((*u2) + PRIME - myModMult(q, v2));
    t3 = myMod((*u3) + PRIME - myModMult(q, v3));
    (*u1) = v1;
    (*u2) = v2;
    (*u3) = v3;
    v1 = t1;
    v2 = t2;
    v3 = t3;
  }while(v3 != 0 && v3!= PRIME);
}

//Computes the modular multiplicative inverse of a modulo m,
//using the extended Euclidean algorithm
//only works for p=2^61-1
uint64 inv(uint64 a)
{
  uint64 u1;
  uint64 u2;
  uint64 u3;
  extEuclideanAlg(a,&u1, &u2, &u3);
  if(u3==1)
      return myMod(u1);
  else
      return 0;
}
   
//extrapolate vector vec of length n to location r
uint64 extrap1d(uint64* vec, uint64 n, uint64 r)
{
  uint64 result=0;
  uint64 mult=1;
  for(uint64 i = 0; i < n; i++)
  {
    mult=1;
    for(uint64 j=0; j < n; j++)
    {
      if (i>j)
        mult=myModMult(myModMult(mult, myMod(r-j+PRIME)), inv(i-j) );
      if (i<j)
        mult=myModMult(myModMult(mult, myMod(r-j+PRIME)), inv(myMod(i+PRIME-j)) );
    }
    result=myMod(result+myModMult(mult, vec[i]));
  }
  return result;
}

//computes sum message (i.e. a message in a round where a sum operator is being stripped off)
//and modifies mcounts of each item appropriately
uint64* sum_message(uint64 j, full_id* identifiers, uint64 streamlength, 
uint64 d, uint64* r)
{
  uint64* y = (uint64*) calloc(2, sizeof(uint64));
  uint64 zero_count=0;
  uint64 one_count=0;
  uint64 item_mask = (((uint64) 1) << d)-1; 
  uint64 litem_bits = 0; //"item bits" of previous item

  for(uint64 i = 0; i < streamlength; i++)
  {     
    if((identifiers[i].id>>j)&1)
    {
      if( (i == 0) || ((identifiers[i].id&item_mask) != litem_bits))
      {
        one_count = myMod(one_count + identifiers[i].mcount);
      }
      identifiers[i].mcount = myModMult(identifiers[i].mcount, r[j]);
    }
    else
    {
      if( (i == 0) || ((identifiers[i].id&item_mask) != litem_bits))
      {
        zero_count = myMod(zero_count + identifiers[i].mcount);
      }
      identifiers[i].mcount = myModMult(identifiers[i].mcount, 1+PRIME-r[j]);
    }
    litem_bits = identifiers[i].id&item_mask;
  }

  y[0] = zero_count;
  y[1] = one_count;

  return y;
}

//computes linearization message (i.e. a message in a round where a linearization operator is being stripped off)
//and modifies mcounts of each item appropriately
uint64* lin_message(uint64 round_num, uint64* r, uint64 d, 
uint64 var, uint64* last_bound, uint64 next_or, full_id* identifiers, 
uint64 streamlength)
{
  //P's message is a degree-two polynomial because of stripped linearization
  uint64* y = (uint64*) calloc(3, sizeof(uint64));
  uint64 inv_r_j = inv(r[last_bound[var]]);
  uint64 inv_1_min_r_j = inv(1+PRIME-r[last_bound[var]]);

  uint64 bound_p_next_or_mask = (((uint64) 1)<<(next_or+1))-1; //pulls out bound vars, including next-or bit
  uint64 next_or_mask = ((uint64) 1)<<next_or; //pulls out variable next_or
  uint64 lin_mask = (((uint64) 1)<<var)-1; //pulls out bits that have been linearized
  uint64 lbound_p_next_or_bits = 0;

  uint64 zero_val[3]; //(OR f)(x, y) = f(x, 0) + f(x, 1) - f(x,0) * f(x,1). This is f(x, 0) for the appropriate f
  uint64 one_val[3]; //f(x, 1) for the appropriate f

  for(uint64 j = 0; j < 3; j++)
  {
     zero_val[j]=0;
     one_val[j]=0;
  }

  int move_on = 0;

  for(uint64 i = 0; i < streamlength; i++)
  {
    if((identifiers[i].id>>var)&1) //if var being unbound is 1
    {
      //first unbind variable var
      identifiers[i].mcount = myModMult(identifiers[i].mcount, inv_r_j);
      //inv_lins tracks inverse of product of values of linearized variables
      if(var == next_or-1)
	 identifiers[i].inv_lin = inv(identifiers[i].mcount);
      else
         identifiers[i].inv_lin = myModMult(identifiers[i].inv_lin, r[last_bound[var]]);
      
      //make sure P ``skips" items as ORs require
      if((i == 0) || (next_or == (2*d-1)) || (identifiers[i].id & bound_p_next_or_mask) !=  lbound_p_next_or_bits)
      {
        y[1] = myMod(y[1]+identifiers[i].mcount);
        y[2] = myMod(y[2]+2*identifiers[i].mcount); 
              
        if((identifiers[i].id & next_or_mask) != 0)
        {
          one_val[1] = myMod(one_val[1] + identifiers[i].mcount);
          one_val[2] = myMod(one_val[2] + 2*identifiers[i].mcount);
        }
        else
        {
          zero_val[1] = myMod(zero_val[1] + identifiers[i].mcount);
          zero_val[2] = myMod(zero_val[2] + 2* identifiers[i].mcount);
        }
      }
      //rebind variable var for this item
      identifiers[i].mcount = myModMult(identifiers[i].mcount, r[round_num]);
    }
    else
    {
      identifiers[i].mcount = myModMult(identifiers[i].mcount, inv_1_min_r_j);
      if(var == next_or-1)
	identifiers[i].inv_lin = inv(identifiers[i].mcount);
      else
        identifiers[i].inv_lin = myModMult(identifiers[i].inv_lin, 1+PRIME-r[last_bound[var]]);
      if((i == 0) || (next_or == (2*d-1)) || (identifiers[i].id & bound_p_next_or_mask) !=  lbound_p_next_or_bits)
      {
        y[0] = myMod(y[0]+identifiers[i].mcount);
        y[2] = myMod(y[2]+myModMult((PRIME-1), identifiers[i].mcount));
        
        if((identifiers[i].id & next_or_mask) != 0)
        {
          one_val[0] = myMod(one_val[0] + identifiers[i].mcount);
          one_val[2] = myMod(one_val[2] + myModMult((PRIME-1), identifiers[i].mcount));
        }
        else
        {
          zero_val[0] = myMod(zero_val[0] + identifiers[i].mcount);
          zero_val[2] = myMod(zero_val[2] + myModMult((PRIME-1), identifiers[i].mcount));
        }
      }   
      identifiers[i].mcount = myModMult(identifiers[i].mcount, 1+PRIME-r[round_num]);
    }
    if( (i < (streamlength-1)) && (identifiers[i].id & lin_mask) != (identifiers[i+1].id & lin_mask))
	  move_on = 1;
    if((move_on == 1) || i == (streamlength-1))
    {
      y[0] = myMod(y[0] +PRIME- myModMult(identifiers[i].inv_lin, myModMult(zero_val[0], one_val[0])));
      y[1] = myMod(y[1] +PRIME- myModMult(identifiers[i].inv_lin, myModMult(zero_val[1], one_val[1])));
      y[2] = myMod(y[2] +PRIME- myModMult(identifiers[i].inv_lin, myModMult(zero_val[2], one_val[2])));
       
      for(uint64 j = 0; j < 3; j++)
      {
        zero_val[j]=0;
        one_val[j]=0;
      }
      move_on=0;
    }
    lbound_p_next_or_bits = (identifiers[i].id & bound_p_next_or_mask);
  }
  return y;
}

//computes an OR message (i.e. a message in a round where an OR operator is being stripped off)
//and modifies mcounts of each item appropriately
uint64* or_message(uint64 j, full_id* identifiers, uint64 streamlength, uint64* r, uint64 d, uint64 var)
{
  //H's message is a degree-one polynomial
  uint64* y = (uint64*) calloc(2, sizeof(uint64));
  uint64 bound_mask = (((uint64) 1) << (var+1))-1; //pulls out all bound variables, including var being unbound
    					       //(essentially we are computing next-or with all cross-terms killed)
    

  uint64 lunbound_bits = 0;


  for(uint64 i = 0; i < streamlength; i++)
  {
    //OR over all variables strictly greater than var (indexed from 0) 
    if( (identifiers[i].id >> var) & 1)
    {
      if( (i == 0) || (var== (2*d-1)) || (identifiers[i].id & bound_mask) != lunbound_bits)
      { 
      	y[1] = myMod(y[1] + identifiers[i].mcount);
      }
      identifiers[i].mcount = myModMult(identifiers[i].mcount, r[j]);
    }
    else
    {
      if( (i == 0) || (var== (2*d-1)) || (identifiers[i].id & bound_mask) != lunbound_bits) 
      {
        y[0] = myMod(y[0] + identifiers[i].mcount);
      }
      identifiers[i].mcount = myModMult(identifiers[i].mcount, 1+PRIME-r[j]);
    }
    lunbound_bits = identifiers[i].id & bound_mask;
  }
  return y;
}

//cycles through the rounds, computing P and V's messages and timing how long it takes
double rounds(uint64 d, uint64* r, double* ct_secs, uint64 final_check, full_id* identifiers, uint64 streamlength)
{
  uint64 num_rounds = 2*d + ((3*d-1)*d)/2;
  //y will store the prover's messages
  uint64** y = (uint64**) malloc(num_rounds * sizeof(uint64*));

  int* round_type = (int*) calloc(num_rounds, sizeof(int));
  int* round_var = (int*) calloc(num_rounds, sizeof(int));
  int lins_to_do=d;

  int count = 0;
  //one iteration for each sum op and OR op
  for(uint64 i = 0; i < 2*d; i++)
  {
     if(i < d)
     {
	round_type[count] = 0; //0 means sum operation
        round_var[count]=count;
        count++;
        continue;
     }
     //we're on an OR operation. Need to do a bunch of linearizations before we can do OR
     for(uint64 j = 0; j < lins_to_do; j++)
     {
        round_type[count] = 1; //1 means lin operation
        round_var[count] = i-j-1;
        count++;
     }
     lins_to_do++;
     round_type[count] = 2; //2 means OR operation
     round_var[count]=i;
     count++;
  }

  if(count != num_rounds)
  {
	cout << "there's a bug. Count != num_rounds. num_rounds: " << num_rounds << "count: " << count << endl;
	exit(1);
  }

  uint64 check=0;
  uint64 ct=0; //check time
  uint64 k = 0;
  //tracks for each of the 2d variables x_i the last round in which x_i was bound
  uint64* last_bound = (uint64*) calloc(2*d, sizeof(uint64));
  uint64 next_or = d;

  clock_t t = clock();

  //In a real system, P and V's messages need to be interleaved so
  //P doesn't know r_j until after he commits to g_j. But here, we are
  //just interested in determining how long it takes to compute the messages.

  for(uint64 j=0; j < num_rounds; j++) // computes all the messages from prover to verifier
  {
      if(round_type[j] == 0)
      {
        last_bound[j]=j;
        y[j] = sum_message(j, identifiers, streamlength, d, r);
      }
      if(round_type[j] == 1)
      {
         y[j] = lin_message(j, r, d, round_var[j], last_bound, next_or, identifiers, streamlength);
         last_bound[round_var[j]] = j;
      }
      if(round_type[j] == 2)
      {
         y[j] = or_message(j, identifiers, streamlength, r, d, round_var[j]); 
         last_bound[round_var[j]]=j;
	 next_or++;
      }        	
  }
  clock_t pt=clock()-t; //prover time

  cout << "claimed F0 is: " << myMod(y[0][0] + y[0][1]) << endl;
  uint64 last;

  t=clock();
  last_bound[0]=0;
  for(uint64 j = 0; j < num_rounds; j++) //checks all messages from prover to verifier
  {
    if (j>0 && round_type[j] == 0)
    {
    //  cout << "setting last_bound[" << j << "] to: " << j << endl;
      last_bound[j]=j;
      if( (check != myMod(y[j][0] + y[j][1])) && (check != myMod(y[j][0] + y[j][1]) + PRIME)
          && (check + PRIME != myMod(y[j][0] + y[j][1])) ) // check consistency of messages ie y[j][0]+y[j][1] = check
      {
        cout << "Check failed in sum round: j is " << j << "check is: ";
	cout << check << "!=" <<  myMod(y[j][0] + y[j][1]);
        cout << "y[" << j << "][0]: " << y[j][0] << " y[" << j << "][1]: " << y[j][1] << endl;
	cout << "y[" << j << "][0]: " << (signed long) (y[j][0]- PRIME);
	cout  << " y[" << j << "][1]: " << (signed long) (y[j][1]- PRIME) << endl; 
	cout << "check " << (signed long) (check- PRIME) << endl;
      }
    }

    if(round_type[j]==1)
    {
      last= last_bound[round_var[j]];
      //if i is last time var we are unbinding was bound, check that r_i * y[j][1] + (1-r_i) * y[j][0] = check
      if( (myMod(myModMult(r[last], y[j][1]) + myModMult(1+PRIME-r[last], y[j][0]) + PRIME - check) != 0) && 
          (myMod(myModMult(r[last], y[j][1]) + myModMult(1+PRIME-r[last], y[j][0]) + PRIME - check) != PRIME)) 
      {
        cout << "Check failed in linearization round: j is " << j << "check is: ";
	cout << check << "!=" <<  myMod(myModMult(r[last], y[j][1]) + myModMult(1+PRIME-r[last], y[j][0]));
        cout << "y[" << j << "][0]: " << y[j][0] << " y[" << j << "][1]: " << y[j][1] << endl;
	cout << "y[" << j << "][0]: " << (signed long) (y[j][0]- PRIME);
	cout  << " y[" << j << "][1]: " << (signed long) (y[j][1]- PRIME) << endl; 
	cout << "check " << (signed long) (check- PRIME) << endl;
        cout << "last is: " << last << "r[last] is: " << r[last] << endl;
        cout << "round_var[j] is: " << round_var[j] << endl;
      }
      //indicate this is the most recent round in which j was bound
      last_bound[round_var[j]] = j;
    }

    if(round_type[j]==2)
    {
      if( (myMod(y[j][0] + y[j][1] + PRIME - myModMult(y[j][0], y[j][1]) + PRIME - check) != 0) && 
          (myMod(y[j][0] + y[j][1] + PRIME - myModMult(y[j][0], y[j][1]) + PRIME - check) != PRIME)) 
      {
        cout << "Check failed in OR round: j is " << j << "check is: ";
	cout << check << "!=" <<  myMod(y[j][0] + y[j][1] + PRIME - myModMult(y[j][0], y[j][1]));
        cout << "y[" << j << "][0]: " << y[j][0] << " y[" << j << "][1]: " << y[j][1] << endl;
	cout << "y[" << j << "][0]: " << (signed long) (y[j][0]-PRIME);
	cout  << " y[" << j << "][1]: " << (signed long) (y[j][1]-PRIME) << endl; 
	cout << "check " << (signed long) (check- PRIME) << endl;
      }
      //indicate this is the most recent round in which j was bound
      last_bound[round_var[j]] = j;
    }

    if((round_type[j] == 0) || (round_type[j] == 2)) 
    	check=extrap1d(y[j], 2, r[j]); //compute next check value at random location r
    else
	check = extrap1d(y[j], 3, r[j]);
  }

  if((check != final_check) && (check+PRIME != final_check) && (check != final_check + PRIME))
  {
 	cout << "Final check failed!" << endl;
        cout << "check: " << check << "final_check: " << final_check << endl;
        cout << "dif is: " << myMod(check+PRIME-final_check) << endl;
  }

  *ct_secs = ((double) ct)/CLOCKS_PER_SEC;
  free(round_type);
  free(round_var);
  free(last_bound);
  for(int i = 0; i < num_rounds; i++)
    free(y[i]);
  free(y);
  return ((double) pt)/CLOCKS_PER_SEC; 
}

int main(int argc, char** argv)
{  
  std::cout.setf(ios::fixed, ios::floatfield);
  std::cout.setf(ios::showpoint);
  std::cout.precision(10);

  uint64 d; 

  if(argc < 2) 
  {
    cout << "not enough command line args\n";
    exit(1);
  }

  d=atoi(argv[1]); //read dimensionality from command line
  if(d > 32)
  {
     cout << "We require d <= 32 so we can represent each item of the universe of size u^2 in a 64-bit long\n";
     exit(1);
  }
  uint64 u=myPow(2, 2*d); //universe size
  uint64 streamlength=myPow(2, d); //length of stream
  
  //a will store the data stream (in these experiments, I am setting stream length equal to universe size)
  uint64* a = (uint64*) calloc(streamlength, sizeof(uint64));
  full_id* identifiers = (full_id*) calloc(streamlength, sizeof(full_id));
  if( (identifiers == NULL) || (a == NULL))
  {
     cout << "must not have enough memory; identifiers is NULL" << endl;
     exit(1);
  }

  uint64 num_rounds = 2*d + ((3*d-1)*d)/2;
  //cout << "num_rounds is: " << num_rounds << endl;
  uint64* r=choose_r(num_rounds);

  //stores the final values that each variable is bound to in the protocol
  uint64* final_r = (uint64*) calloc(2*d, sizeof(uint64));

  for(int k = 0; k <= 2*d-2; k++)
    final_r[k] = r[num_rounds-k-2];
  final_r[2*d-1] = r[num_rounds-1]; 

  clock_t t = clock();;
  //note only the last d words of r represent the final bound values of all variables
  uint64** chilook = make_lookup(final_r, d);
  clock_t table_time = clock() - t;

  uint64 w=8; //block length.
  if(w > 2*d)
   w=2*d;
  uint64 rw=myPow(2, w);
  uint64 rd=2*d/w;//number of blocks of length w we need to look over in the lookup table. 
			//True universe size is 2*d because we are viewing input as matrix here, not as a vector
  if((2*d)%w>0)
    rd+=1;

  uint64 mask=rw-1;

  uint64 i=0;
  srand(time(NULL));
  uint64 F0=0;
  clock_t vt = 0;
  uint64 item;

  uint64 hash=0;
  uint64 count;
  
  i=0;
  
  uint64 identifier;
  t=clock();
  while(i<streamlength) //randomly generate stream
  {
    item = rand() % streamlength;
    identifier = (i << d) + item;
    if(a[item] == 0) 
	F0++;
    a[item]=a[item]+1;
    identifiers[i].id = identifier;
    identifiers[i].mcount=1;
    identifiers[i].inv_lin=1;
    //handle V's processing of the stream (Extrapolation)
    update_hash(&hash, chilook, rd, w, rw, mask, identifier);
    i+=1;
  }
  vt += clock()-t;

  t=clock();
  qsort(identifiers, streamlength, sizeof(full_id), compare);
  clock_t sort_t = clock()-t;
  cout << "sort time is: " << ((double) sort_t)/CLOCKS_PER_SEC << endl;

  cout << "Actual F0 is: " << F0 << endl;
  double vt_secs =((double) vt)/CLOCKS_PER_SEC;

  double ct_secs;

  double pt_secs=rounds(d, r, &ct_secs, hash, identifiers, streamlength);
   
  cout << "N\tVerifT\tProveT\tCheckT\tVerifS\tProofS \n";
  cout << myPow(2, d) << "\t" << vt_secs << "\t" << pt_secs << "\t" << ct_secs << "\t" << d+1 << "\t" << 3*d+1 << endl;

  //clean up memory before exiting
  free(final_r);
  free(r);
  free(identifiers);
  for(int i = 0; i < rd; i++)
    free(chilook[i]);
  free(chilook);
  free(a);
}
