/*****************************
//Justin Thaler
//May 7, 2011
//Implementation of interactive F2 protocol, originally due 
//to "Verifying Computations with Streaming Interactive Proofs"
//by Cormode, Thaler, and Yi.
********************************/

#include <cstdlib>
#include <iostream>
#include <time.h>

#define MASK 4294967295 //2^32-1
#define PRIME 2305843009213693951 //2^61-1

typedef unsigned long long uint64;

using namespace std;

uint64 myPow(uint64 x, int b) 
{
  uint64 i = 1;
  for (int j = 0; j < b; j++) 
      i *= x;
  return i;
}

//efficient modular arithmetic function for p=2^61-1. Only works for this value of p.
//This function might
//return a number slightly greater than p (possibly by an additive factor of 8);
//It'd be cleaner to check if the answer is greater than p
//and if so subtract p, but I don't want to pay that
//efficiency hit here, so the user should just be aware of this.
uint64 myMod(uint64 x)
{
  return (x >> 61) + (x & PRIME);
}  

//efficient modular multiplication function mod 2^61-1
uint64 myModMult(uint64 x, uint64 y)
{
	uint64 hi_x = x >> 32; 
	uint64 hi_y = y >> 32;
	uint64 low_x = x & MASK;
	uint64 low_y = y & MASK;

	//since myMod might return something slightly large than 2^61-1,  
	//we need to multiply by 8 in two pieces to avoid overflow.
	uint64 piece1 = myMod((hi_x * hi_y)<< 3);
	uint64 z = (hi_x * low_y + hi_y * low_x);
	uint64 hi_z = z >> 32;
	uint64 low_z = z & MASK;

	//Note 2^64 mod (2^61-1) is 8
	uint64 piece2 = myMod((hi_z<<3) + myMod((low_z << 32)));
	uint64 piece3 = myMod(low_x * low_y);
	uint64 result = myMod(piece1 + piece2 + piece3);

	return result;
}  

uint64* choose_r(uint64 d)
{
  //choose a random vector of dimension d (this is the random point in the LDE)
  uint64* r = (uint64*) malloc(d * sizeof(uint64));
  for(int i=0; i < d; i++)
  {
     r[i] = myMod(rand()); 
  }  
  return r;
}

//Computes the binary characteristic vector function evaluated over dimensions
//e to d, based on binary index v (view v as a binary vector even though it's an integer), 
//evaluated on vector r. Note this function actually modified v
uint64 chi(uint64 e, uint64 d, uint64 v, uint64* r)
{
  uint64 c = 1;
  for(uint64 i = e; i <d; i++)
  {
    if( v&1 )
      c=myModMult(c, r[i]);
    else
      c=myModMult(c, (1+ myMod(PRIME-r[i])));
    v=v>>1;
  }
  return c;
}

// builds a look up table of products of chi values
// our table will give values for products of blocks of w variables -- 
// this keeps table small (only d/w rows and 2^w cols)
uint64** makechi(uint64 w, uint64 rd, uint64 d, uint64* r)
{
  uint64 rw=myPow(2, w);
  uint64** chilook = (uint64**) malloc(rd * sizeof(uint64*));
  for(int i=0; i < rd; i++)
	chilook[i]=(uint64*) malloc(rw * sizeof(uint64));
  uint64 lo=0;
  uint64 hi=w;
  for(uint64 j = 0; j < rd; j++)
  { //for each block of w coordinates
    for(uint64 i = 0; i < rw; i++)
    { //for each possible set of values of the w coordinates,
      //evaluate chi_1... chi_{2^w} (viewing each subscript as a w-dimensional vector) 
      //over dimensions lo to hi at location vector r
      chilook[j][i]=chi(lo,hi,i,r);
    }
    lo=hi;
    hi=hi+w;
    if (hi>d) //d is the dimensionality of our view of universe [2]^d
      hi=d; //if last block contains fewer than w variables, make sure we only look at ones that exist
  }
  return chilook;
}

//extrapolate the global vector a of length u to location r (a is frequency vector)
//That is, compute \sum_i a_i chi_i(r)
uint64 extrap(uint64* r, uint64 d, uint64 u, uint64* a)
{
  uint64 f=0;
  uint64 w=8; //block length. Before, we had w=d/2 resulting in table of size 2^(d/2)=sqrt(u), which seems like cheating.
  uint64 rw=myPow(2, w);
  uint64 mask=rw-1; //used to pull off last w bits of numbers
  uint64 rd=d/w;//number of blocks of length w we need to look over in the lookup table
  if(d%w>0)
    rd+=1;
  uint64** chilook=makechi(w,rd, d, r);
  
  //temporary variables used below
  uint64 i =0;
  uint64 j=0;
  uint64 delta;
  uint64 k;
  
  while(i<u)
  {
    delta=a[i];
    k=i;
    j=0;
    //evaluate a_i chi_i(r)
    while(j<rd) 
    { 
      delta=myModMult(delta, chilook[j][k&mask]);
      k=k>>w;
      j+=1;
    }
    f = myMod((f+delta));
    i+=1;
  }
  for(int i = 0; i < rd; i++)
    free(chilook[i]);
  free(chilook);
  return f;
}

//central part of the chi function evaluated at dimension j given binary v
uint64 minchi(uint64 j, uint64 v, uint64 c)
{
  if(((v>>j) & 1)==1) //pull out LEAST significant bit of v and see if it is odd
    return c;
  else
    return myMod((1+ PRIME - c));
}

//compute g_j(c) for c=0, 1, 2; g_j is prover's j'th message
//fast computation of the check, by evaluating the implicit poly at 0,1,2
//giving result[0], result[1], result[2]
//This function specfic to second frequency moment.
uint64* fcompgjc(uint64 d, uint64 j, uint64* a)
{
  //Note when this function is called,a[k] is expected to equal
  //A_j[vj...vd]=A_j[k]
  uint64* result = (uint64*) calloc(3, sizeof(uint64));

  uint64 top=myPow(2, (d-j));
  uint64 k=0;
  uint64 val;
  while(k<top)
  {
    result[0]= myMod((result[0] + myModMult(a[k], a[k])));
    result[1]= myMod((result[1] + myModMult(a[k+1], a[k+1])));
    val=myMod(myMod(2*a[k+1])+myMod(PRIME-a[k])); //because chi_1(2)=2 and chi_0(2)=-1
    result[2]= myMod(result[2] + myModMult(val, val));
    k= k+2;
  }
  return result;
}

//slower way to compute the j'th message of F2 protocol (evaluated at c) i.e.
//computes the sum of squared values of the poly over given range 
//-- not using final optimization in paper.
uint64 gjc(uint64 d, uint64 j, uint64 c, uint64* a)
{ 
  uint64 outer=myPow(2, d-j-1); //number of values to sum over in outer sum
  uint64 inner=myPow(2, j+1);  //number of values to sum over in inner sum. 
  uint64 result=0;
  uint64 val = 0;
  uint64 v=0;
  for(uint64 i = 0; i < outer; i++)
  {
    val=0;
    for(uint64 k=0; k < inner; k++)
    {
      val = myMod((val + myModMult(a[v], minchi(j, v,c))));
      v=v+1;
    }
    result=myMod((result + myModMult(val, val)));//we have one point for each outer term. See top of page 7 of paper
		           //i.e. we are computing:
		           // \sum_{v{j+1}...v_d} (\sum_{v_1...v_j} a_v (prod_{k=1}^j-1 chi_{v_k}(r_k)0 chi_{v_j}(c))^2.
		           //Notice a[v] has been multiplied by prod_{k=1}^j-1 chi_{v_k}(r_k) by the time this function is
                           //called, via external calls to the function update_a
  }
  return result;
}

//update the global vector a by adding on the next level of chi (detailed in paper)
//Note this actually modifies a
void update_a(uint64 d, uint64 j, uint64 rj, uint64* a)
{
  for(uint64 i=0; i < myPow(2, d); i++)
    a[i] = myModMult(a[i], minchi(j,i,rj));
}

//update and compact the global vector by pairing up adjacent values and collapsing
//Note this actually modifies a
void compact_a(uint64 d, uint64 j, uint64 rj, uint64* a)
{
  uint64 k=0;
  uint64 i=0;
  uint64 top=myPow(2, d-j-1);
  while(i<top)
  {
    a[i] = myMod(a[k]+ myModMult(rj, myMod( a[k+1]+myMod(PRIME-a[k])))); 
    //Note chi_1(rj)=rj, chi_0(rj)=1-rj.
    i=i+1;
    k+=2;
  }
}

//evaluate TWICE quadratic given by f(0), f(1), f(2) to f(r) (twice to make everything integral, no inverses needed)
uint64 interp(uint64* y, uint64 rj)
{
  uint64 term_one = myModMult( myModMult(rj-1, rj-2), y[0]);
  uint64 term_two = myModMult( myMod(2*myModMult(rj, rj-2)), y[1]);
  uint64 term_three = myModMult(myModMult(rj, rj-1), y[2]);
  return myMod(term_one + (PRIME-term_two) + term_three);
}

//cycles through the rounds, computing P and V's messages and timing how long it takes
double rounds(uint64 d, uint64* r, uint64 compact, uint64* a, uint64 fr, uint64 f2, double* ct_secs)
{
  uint64** y = (uint64**) malloc(d * sizeof(uint64*));
  uint64 check=0;
  uint64 ct=0; //check time
  uint64 k = 0;
  clock_t t = clock();

  //In a real system, P and V's messages need to be interleaved so
  //P doesn't know r_j until after he commits to g_j. But here, we are
  //just interested in determining how long it takes to compute the messages.

  for(uint64 j=0; j < d; j++) // computes all the messages from prover to verifier
  {
    if (compact)
    {
      y[j]=fcompgjc(d,j, a);
      compact_a(d,j,r[j], a);
    }
    else
    {
      y[j]=(uint64*) malloc(3 * sizeof(uint64));
      for(uint64 i = 0; i < 3; i++)
      {
        k=i;
        y[j][i]=gjc(d,j,k, a);
      }
      update_a(d,j,r[j], a);
    }
  }
  clock_t pt=clock()-t; //prover time

  if(f2 != y[0][0] + y[0][1])
    cout << "Claimed f2 does not equal real f2!\n";

  t=clock();
  for(uint64 j = 0; j < d; j++) //checks all messages from prover to verifier
  {

    if (j>0)
    {
      if (  (myMod( (myMod(2* myMod(y[j][0]+y[j][1]))  + myMod(PRIME-check))) != 0)
         && (myMod( (myMod(2* myMod(y[j][0]+y[j][1]))  + myMod(PRIME-check))) !=PRIME) )//check message consistency
      {
        cout << "Check failed: j is " << j << "check is: ";
	cout << check << "!=" << (2*((y[j][0]+y[j][1]))) << "mod p" << endl;
      }
    }
    check=interp(y[j],r[j]); //compute next check value at random location r
  }
  if ( ( ( ( (myModMult(2*fr, fr))+ myMod(PRIME-check))) != 0) && ( ( ( (myModMult(2*fr, fr))+ myMod(PRIME-check)))) != PRIME) // compare to extrapolation of f at r
    cout << "Final check failed, "<< check <<"!=" << 2*fr*fr << "mod p" << endl;
  ct=clock()-t;
  *ct_secs = ((double) ct)/CLOCKS_PER_SEC;

  for(int i = 0; i < d; i++)
    free(y[i]);
  free(y);

  return ((double) pt)/CLOCKS_PER_SEC; 
}

int main(int argc, char** argv)
{  
  std::cout.setf(ios::fixed, ios::floatfield);
  std::cout.setf(ios::showpoint);
  std::cout.precision(10);


  uint64 d; 

  if(argc < 2) 
  {
    cout << "not enough command line args\n";
    exit(1);
  }

  d=atoi(argv[1]); //read dimensionality from command line
  if(d < 8)
  {
    cout << "For simplicity of implementation, d is required to be at least 8. Setting d to 8.\n";
    d=8;
  }
  uint64 u=myPow(2, d); //universe size
  uint64 f2=0;
  
  uint64* a = (uint64*) malloc(u*sizeof(uint64));
  if(a == NULL)
  {
     cout << "must not have enough memory; a is NULL" << endl;
     exit(1);
  }
  uint64 i=0;
  srand(time(NULL));
  while(i<u) // randomly create a vector of values
  {
    a[i] = rand() % 1000;  
    i+=1;
  }

  clock_t temp = clock();
  i=0;
  while(i<u) // randomly create a vector of values
  {
    f2= f2 + a[i]*a[i]; //compute exact F2 as a check
    i+=1;
  }

  clock_t gen = clock()-temp;
  //cout << "Exact F2 computing time " <<  ((double) gen)/CLOCKS_PER_SEC << endl;

  //cout << "exact f2 is " << f2 << endl;

  uint64* r=choose_r(d);
  clock_t t=clock();

  uint64 fr=extrap(r,d, u, a); //evaluate f extrapolated to random location r

  clock_t vt=clock()-t;

  double vt_secs =((double) vt)/CLOCKS_PER_SEC;

  //cout << "Initial verifier time: " <<  ((double) vt)/CLOCKS_PER_SEC << endl;
  double ct_secs;
  double pt_secs=rounds(d, r, 1, a, fr, f2, &ct_secs);
  cout << "N\tVerifT\tProveT\tCheckT\tVerifS\tProofS \n";
  cout << u << "\t" << vt_secs << "\t" << pt_secs << "\t" << ct_secs << "\t" << d+1 << "\t" << 3*d+1 << endl;
  
  //clean up memory before exiting
  free(r);
  free(a);
}
