/*****************************
//Justin Thaler
//May 7, 2011
//Naive Implementation of non-interactive F2 protocol, originally due 
//to "Annotations in Data Streams" by Chakrabarti, Cormode, and McGregor.
//Code developed for the paper
//"Practical Verified Computation with Streaming Interactive Proofs"
//by Cormode, Mitzenmacher, and Thaler.
********************************/

#include <cstdlib>
#include <iostream>
#include <time.h>

#define MASK 4294967295 //2^32-1
#define PRIME 2305843009213693951 //2^61-1

typedef unsigned long long uint64;

using namespace std;

uint64 myPow(uint64 x, int e) 
{
  uint64 i = 1;
  for (int j = 0; j < e; j++)  
     i *= x;
  return i;
}

//efficient modular arithmetic function for p=2^61-1. Only works for this value of p.
//This function might
//return a number slightly greater than p (possibly by an additive factor of 8);
//It'd be cleaner to check if the answer is greater than p
//and if so subtract p, but I don't want to pay that
//efficiency hit here, so the user should just be aware of this.
uint64 myMod(uint64 x)
{
  return (x >> 61) + (x & PRIME);
}  

//efficient modular multiplication function mod 2^61-1
uint64 myModMult(uint64 x, uint64 y)
{
	uint64 hi_x = x >> 32; 
	uint64 hi_y = y >> 32;
	uint64 low_x = x & MASK;
	uint64 low_y = y & MASK;

	//since myMod might return something slightly large than 2^61-1,  
	//we need to multiply by 8 in two pieces to avoid overflow.
	uint64 piece1 = myMod((hi_x * hi_y)<< 3);
	uint64 z = (hi_x * low_y + hi_y * low_x);
	uint64 hi_z = z >> 32;
	uint64 low_z = z & MASK;

	//Note 2^64 mod (2^61-1) is 8
	uint64 piece2 = myMod((hi_z<<3) + myMod((low_z << 32)));
	uint64 piece3 = myMod(low_x * low_y);
	uint64 result = myMod(piece1 + piece2 + piece3);

	return result;
}  

//Performs Extended Euclidean Algorithm
//Used for computing multiplicative inverses mod p
//Computes a solution  to u*u1 + p*u2 = gcd(u,v)=u3
//only works for p=2^61-1
void extEuclideanAlg(uint64 u, uint64* u1, uint64* u2, uint64* u3)
{
  *u1 = 1;
  *u2 = 0;
  *u3 = u;
  uint64 v1 = 0;
  uint64 v2 = 1;
  uint64 v3 = PRIME;
  uint64 q;
  uint64 t1;
  uint64 t2;
  uint64 t3;
  do
  {
    q = *u3 / v3;
    t1 = myMod((*u1) + PRIME - myModMult(q, v1));
    t2 = myMod((*u2) + PRIME - myModMult(q, v2));
    t3 = myMod((*u3) + PRIME - myModMult(q, v3));
    (*u1) = v1;
    (*u2) = v2;
    (*u3) = v3;
    v1 = t1;
    v2 = t2;
    v3 = t3;
  }while(v3 != 0 && v3!= PRIME);
}

//Computes the modular multiplicative inverse of a modulo m,
//using the extended Euclidean algorithm
//only works for p=2^61-1
uint64 inv(uint64 a)
{
  uint64 u1;
  uint64 u2;
  uint64 u3;
  extEuclideanAlg(a, &u1, &u2, &u3);
  if(u3==1)
      return myMod(u1);
  else
      return 0;
}

//build a look-up table of multiplicative inverses
//only works for p=2^61-1
uint64* geninv(uint64 n)
{
  uint64* ifact_plus = (uint64*) calloc(n, sizeof(uint64));
  uint64* ifact_minus = (uint64*) calloc(n, sizeof(uint64));
  ifact_plus[0]=1;
  ifact_minus[0]=1;
  uint64 i;
  for(i=1; i < n; i++)
	ifact_plus[i] = myModMult(inv(i), ifact_plus[i-1]);  
  for(i=1; i < n; i++)
	ifact_minus[i] = myModMult(inv(-i+PRIME), ifact_minus[i-1]); 
  uint64* table= (uint64*) calloc(n, sizeof(uint64));
  for(i=0; i < n; i++)
	table[i] = myModMult(ifact_plus[i], ifact_minus[n-i-1]);
  free(ifact_plus);
  free(ifact_minus);
  return table;
}	

//deprecated because this is a much slower way to compute itab
//build a look-up table of multiplicative inverses
//only works for p=2^61-1
uint64* geninvOLD(uint64 n)
{
  uint64* table = (uint64*) calloc(n, sizeof(uint64));
  uint64 mult=1;
  for(uint64 i = 0; i < n; i++)
  {
    mult=1;
    for(uint64 j=0; j < n; j++)
    {
      if(i>j)
      {
        mult=myModMult(mult, inv(myMod(i-j)));
      }
      if(i < j)
      {
        mult=myModMult(mult, inv(myMod((PRIME+i)-j)));
      }
    }
    table[i]=mult;
  }
  return table;
}


//deprecated (as long as r >=n, which happens with extremely high probability)
//even if r<n there are faster ways to generate the lookup table, but this never
//happens so I'm not bothering to deal with it.
//generate a look up table of factorials needed for lagrange polys
//i'th entry is numerator of chi_x(r)
uint64* genrtabOLD(uint64 n, uint64 r)
{
  uint64* table = (uint64*) calloc(n, sizeof(uint64));
  uint64 mult=1;
  
  uint64* lookup = (uint64*) calloc(n, sizeof(uint64));
  for(uint64 i = 0; i < n; i++)
	lookup[i] = myMod(r-i+PRIME);
  for(uint64 i = 0; i < n; i++)
  {
    mult=1;
    for(uint64 j=0; j < n; j++)
    {
      if(i != j)
        mult=myModMult(mult, lookup[j]);
    }
    table[i]=mult;
  }
  free(lookup);
  return table;
}

//generate a look up table of factorials needed for lagrange polys
//i'th entry is numerator of chi_x(r)
uint64* genrtab(uint64 n, uint64 r)
{
  if(r < n)
	return genrtabOLD(n, r);
  
  uint64* table = (uint64*) calloc(n, sizeof(uint64));
  uint64 mult=1;
  
  //I will use the fact that rtab[i] = rtab[i-1](r-i+1)/(r-i) if r != i, r!=i-1
  for(uint64 j=1; j < n; j++)
  {
      mult=myModMult(mult, r-j);
  }
  table[0]=mult;
    
  for(uint64 i = 1; i < n; i++)
  {
    table[i]=myModMult(myModMult(table[i-1], r-i+1), inv(r-i));
  }
  return table;
}

//extrapolate vector vec of length n to location r
//very slow, never used
uint64 extrap(uint64* vec, uint64 n, uint64 r)
{
  uint64 result=0;
  uint64 mult=1;
  for(uint64 i = 0; i < n; i++)
  {
    mult=1;
    for(uint64 j=0; j < n; j++)
    {
      if (i>j)
        mult=myModMult(myModMult(mult, myMod(r-j+PRIME)), inv(i-j));
      if (i<j)
        mult=myModMult(myModMult(mult, myMod(r-j+PRIME)), inv(myMod(i+PRIME-j)));
    }
    result=myMod(result+myModMult(mult, vec[i]));
  }
  return result;
}

//extrapolate vector vec of length n to location r using lookup table of inverses itab
//but not a lookup table of factorials. Also slow, never used.
uint64 medextrap(uint64* vec,uint64 n, uint64 r, uint64* itab)
{
  uint64 result=0;
  uint64 mult=1;

  uint64* lookup = (uint64*) calloc(n, sizeof(uint64));
  for(uint64 i = 0; i < n; i++)
	lookup[i] = myMod(r-i+PRIME);

  for(uint64 i = 0; i < n; i++)
  {
    mult=1;
    for(uint64 j=0; j < n; j++)
    {
      if (i!=j)
        mult=myModMult(mult, lookup[j]);
    }
    //cout << "i is\n" << i << endl;
    result=myMod(result+ myModMult(myModMult(mult, itab[i]), vec[i]));
  }
  free(lookup);
  return result;
}

//extrapolate vector vec of length n to location r
//using lookup table of inverses itab and of factorials rtab
uint64 tabextrap(uint64* vec, uint64 n, uint64 r, uint64* itab, uint64* rtab)
{
  //we never use r here?
  uint64 result=0;
  for(uint64 i=0; i< n; i++)
  {
     result=myMod(result+myModMult(myModMult(rtab[i], itab[i]), vec[i]));
  }     
  return result;
}

//randomly generate data of size v*h and compute exact F2
uint64 gendata(uint64 h, uint64 v, uint64** a)
{
  uint64 f2=0;
  for(uint64 i=0; i< v; i++)
  {
    for(uint64 j=0; j< h; j++)
    {
      a[i][j]=rand() % 1000;
      f2+=a[i][j]*a[i][j];
    }
  }
  return f2;
}

//generate the check values for the verifier by extrapolating each row to column r
uint64 vcheck(uint64 h, uint64 v, uint64 r, uint64** a, uint64* itab)
{
  uint64 check=0;
  uint64* rtab=genrtab(h, r);
  uint64 ext;
  for(uint64 i=0; i < v; i++)
  {
    ext=tabextrap(a[i],h,r,itab,rtab); //compute LDE(a)(i, r)
    check=myMod(check+myModMult(ext, ext)); //in end, check = sum_i=1^v LDE(a)(i,r)^2
  }
  free(rtab);
  return check;
}

//build the proof by extrapolating the data to adjacent columns, 
//then summing to get a vector of sums of squares, which can be extrapolated by V
uint64* proof(uint64 h, uint64 v, uint64** facttab, uint64* itab, uint64** a)
{
  uint64 i;
  uint64 j;
  for(i = 0; i < v; i++)
  {
    for(j=h; j<2*h; j++)
    {
      a[i][j]=tabextrap(a[i],h,j,itab,facttab[j-h]);
    }
  }
  uint64* b=(uint64*) calloc(2*h, sizeof(uint64));   
  uint64 f = 0;
  for(j=0; j<2*h; j++)
  {
    f=0;
    for(i=0; i<v; i++)
    {
      f=myMod(f+myModMult(a[i][j], a[i][j]));
    }
    b[j]=myMod(f);
  }
  return b;
}

int main(int argc, char** argv)
{
    if(argc < 2)
    {
      cout << "too few args\n";
      exit(1);
    }
    uint64 d=(uint64) atoi(argv[1]); //read dimensionality from command line

    uint64 h=d;
    uint64 v=d; //assume square array

    //prepare space for data
    uint64** a = (uint64**) calloc(v, sizeof(uint64*));
    uint64 i;
    for(i=0; i < v; i++)
      a[i] = (uint64*) calloc(2*h, sizeof(uint64));
    i=0;
    uint64 r=rand(); //pick random location r
    uint64 cf2=gendata(h,v, a);

    //have prover generate his lookup tables. We charge him for this even though they
    //are data independent so he could do this in advance (this cost is quite small now anyway)
    clock_t t=clock();
    uint64* itab=geninv(h); //table of denominators of lagrange polynomials
    clock_t itab_time = clock()-t;

    uint64** facttab = (uint64**) calloc(2*h, sizeof(uint64*));

    for(uint64 k = h; k < 2*h; k++)
      facttab[k-h]= genrtab(h,k); //for each x-coordinate in ghost rows
    //generate numerator of lagrange basis polynomail chi_x
    clock_t ptable_time=clock()-t;

    //have verifier generate his lookup tables (he will use itab also, which we created above). 
    //We will charge for these as well even though they are data-independent (cost is small anyway)
    t=clock();
    uint64* itab2=geninv(2*h);
    uint64* rtab2=genrtab(2*h, r);
    uint64 check=vcheck(h,v,r, a, itab); // make verifiers check values
        
    clock_t vt=clock()-t;

    uint64* b;
    clock_t pt;
    clock_t ct;
    uint64 f2;
    uint64 result;
    t=clock();
    b=proof(h,v,facttab, itab, a); //create prover's proof
    pt=clock()-t;
    //itab2=geninv(2*h)
    t=clock();
    result = tabextrap(b, 2*h, r, itab2, rtab2);
    //result=medextrap(b,2*h,r,itab2);
    //result=extrap(b,2*h,r); //extrapolate the proof vector to location r
    ct=clock()-t;

    f2=0;
    for(i= 0; i < h; i++)
      f2+=b[i]; //compute claimed F2 value
	
    if ((result!=check) || (f2!=cf2))
    {
      cout << "FAIL!\n" << "result: " << result << " check: " << check;
      cout << "f2: " << f2 << " cf2: " << cf2 << endl;
    }  
    cout << "N\tVerifT\tProveT\tCheckT\tVerifS\tProofS\n";
    // print a header if requested
    cout << h*v << "\t" << (double) (vt+itab_time)/CLOCKS_PER_SEC << "\t" << (double) (pt+ptable_time)/CLOCKS_PER_SEC;
    cout << "\t" <<  (double) ct/CLOCKS_PER_SEC  << "\t" << v << "\t" <<2*h << " " << endl;
    //print output
	
    //clean up memory before exiting
    free(itab);
    free(itab2);
    free(rtab2);
    free(b);
    for(int i = 0; i < v; i++)
      free(a[i]);
    free(a);
    for(int i = 0; i < h; i++)
      free(facttab[i]);
    free(facttab);
}

